//Graphics.h
//Functions, defines, and structs to work with 320x200 256 color VGA mode 13h.
#include <dos.h>
#include <memory.h>
#include <stdlib.h>
#include <conio.h>
#include <iostream.h>

//The computers at school do not recognize bool
#ifndef bool
#define bool unsigned char
#define true 1
#define false 0
#endif

//Various constants
#define VGA256              0x13
#define TEXT_MODE           0x03
#define VGA_VIDEO_BUFFER    0xA0000000L
#define MAXX                320
#define MAXY                200
#define SCREEN_HEIGHT       320
#define SCREEN_WIDTH        200
#define MAXCOLORS           256
#define SCREENSIZE          MAXX*MAXY
#define PALETTE_MASK        0x3C6
#define PALETTE_REGISTER_RD 0x3C7
#define PALETTE_REGISTER_WR 0x3C8
#define PALETTE_DATA        0x3C9
#define VGA_INPUT_STATUS    0x3DA
#define VGA_VSYNC_MASK      0x08
#define ASCII_TEXT          0xF000FA6EL
#define CHAR_HEIGHT         8
#define CHAR_WIDTH          8

//Macros
#define PAUSE while (!kbhit()) {}; getch()
#define LINEFEED CR(conx,cony);\
                 if(cony+CHAR_HEIGHT >= MAXY) {\
                    Cls();\
                    conx=0;\
                    cony=0;\
                 }



//Colors in the palette are composed of red, green, and blue intensity values.
typedef struct RGB_color_typ {
   unsigned char red;
   unsigned char green;
   unsigned char blue;
} RGB_color, *RGB_color_ptr;

unsigned char far *video_buffer = (char far *)VGA_VIDEO_BUFFER;
unsigned char far *ascii_text = (char far *)ASCII_TEXT;

void Set_Video_Mode(unsigned char mode);
void Cls(void);
void inline SetPixel(int x, int y, unsigned char c);
unsigned char inline GetPixel(int x, int y);
void inline Set_Palette_Register(int index, RGB_color color);
void inline Get_Palette_Register(int index, RGB_color &color);
void V_Line(int x, int starty, int endy, unsigned char color);
void H_Line(int startx, int endx, int y, unsigned char color);
void Rotate_Palette(void);
bool Vertical_Retrace(void);
void Wait_For_Vertical_Retrace(void);
void Blit_Character(int x, int y, unsigned char t, int c, int bgc,
                    bool transparent);
void Blit_Scaled_Character(int x, int y, unsigned char t, int c, int bgc,
                           bool transparent, float xscale, float yscale);
void Blit_String(char * string, int x, int y, int c, int bgc, bool trans);
void Blit_Con_String(char * string, int &x, int y, int c, int bgc, bool trans);
char * Input_String(int x, int y, int c, int bgc, int maxsize);
char * Input_Con_String(int &x, int y, int c, int bgc, int maxsize, bool rot=false);
void Blit_Scaled_String(char * string, int x, int y, int c, int bgc,
                        bool trans, float xscale, float yscale);
void Box(int minx, int miny, int maxx, int maxy, unsigned char color);
void Filled_Box(int minx, int miny, int maxx, int maxy, unsigned char color);

void MemError(void) {
   Set_Video_Mode(TEXT_MODE);
   cout << "Out of memory.\n";
   asm {
      mov ax, 4C00h
      int 21h
   }
   return;
}

void Set_Video_Mode(unsigned char mode) {
   asm {
      mov ah, 0     // Set the video mode subfunction.
      mov al, mode  // Video mode to which to change.
      int 10h       // Call the video interrupt to effect the change.
   }
}

void Cls(void) {
   _fmemset((char far *)VGA_VIDEO_BUFFER,0,MAXX*MAXY+1);
   return;
}

void inline SetPixel(int x, int y, unsigned char c) {
   //If desired pixel is on screen, plot it.
   if (x >= 0 && x < MAXX && y >= 0 && y < MAXY)
   video_buffer[(y<<8)+(y<<6)+x]=c; //Pixel is set by setting a bit in the
                                    //Video buffer. 
}

unsigned char inline GetPixel(int x, int y) {
   return video_buffer[(y<<8)+(y<<6)+x];
}

//Sets a single color value in the palette indicated by 'index' to the color
//specified by 'color'.
void inline Set_Palette_Register(int index, RGB_color color) {
   //Inform the VGA card that we are going to work with the palette
   //register.

   outp(PALETTE_MASK, 0xFF);

   //Tell the VGA card which register we'll be updating.

   outp(PALETTE_REGISTER_WR, index);

   //Now write the RGB triple, each written to the same port.

   outp(PALETTE_DATA, color.red);
   outp(PALETTE_DATA, color.green);
   outp(PALETTE_DATA, color.blue);

   return;
}

//Read the current setting of a specific palette register, return in 'color'.
void inline Get_Palette_Register(int index, RGB_color &color) {
   //Inform the VGA card that we are working with the palette register.
   outp(PALETTE_MASK,0xFF);

   //Tell the VGA card which register to access.
   outp(PALETTE_REGISTER_RD, index);

   //Now read the data.
   color.red   = inp(PALETTE_DATA);
   color.green = inp(PALETTE_DATA);
   color.blue  = inp(PALETTE_DATA);

   return;
}

//Draw a vertical line.
void V_Line(int x, int starty, int endy, unsigned char color) {
   unsigned int line_offset, index;

   if (x >= 0 && x < MAXX && starty >= 0 && starty < MAXY && endy >= 0
       && endy < MAXY && starty < endy) {

   //Compute the staring position. Actually calculates y*256+y*64 which is
   //effectively y*320. The double shift is actually faster than the
   //multiplication would be. This is, however, only valid if MAXX = 320.

      line_offset = ((starty<<8) + (starty<<6)) + x;

      //Draw each pixel of the line.
      for (index=0; index <= endy-starty; index++) {
         video_buffer[line_offset] = color;

         line_offset += MAXX; //Move down one pixel on the screen.
      }
   }

   return;
}

void Rotate_Palette(void) {
   RGB_color temp, temp2;
   int cnt;
   Wait_For_Vertical_Retrace();
   Get_Palette_Register(64, temp);
   for (cnt=64; cnt < 128; cnt++) {
      Get_Palette_Register(cnt+1, temp2);
      Set_Palette_Register(cnt, temp2);
   }
   Wait_For_Vertical_Retrace();
   for (cnt=128; cnt < 191; cnt++) {
      Get_Palette_Register(cnt+1, temp2);
      Set_Palette_Register(cnt, temp2);
   }
   Set_Palette_Register(191, temp);
}
   
//Draw a horizontal line.
void H_Line(int startx, int endx, int y, unsigned char color) {
   //Uses the fmemset function, faster.
   //Format of fmemset is
   //fmemset(pointer_to_start, value_to_fill_with, length_of_section)

   if (startx >= 0 && startx < MAXX && endx >= 0 && endx < MAXX && y >= 0 &&
       y < MAXY && startx < endx)
      _fmemset((char far *)(video_buffer + ((y<<8) + (y<<6)) + startx), color,
              endx-startx+1);
   return;
}

bool Vertical_Retrace(void) {
   //Bit 4 of VGA_INPUT_STATUS register is set during vertical retrace.
   return inp(VGA_INPUT_STATUS) & VGA_VSYNC_MASK;
}

void Wait_For_Vertical_Retrace(void) {
   while (Vertical_Retrace()) {}
   while (!Vertical_Retrace()) {}
}

void Blit_Character(int x, int y, unsigned char t, int c, int bgc,
                    bool transparent) {
   //Uses the ROM 8x8 character set to blit a character on the screen.

   unsigned char far *work_char;
   unsigned char bit_mask;
   int xx, yy;

   //Compute starting offset into table.
   work_char = ascii_text + (t << 3);

   for (yy=y; yy < y + CHAR_HEIGHT; yy++) {
      bit_mask = 0x80;
      for (xx=x; xx < x + CHAR_WIDTH; xx++) {
         if (bit_mask & *work_char) {
            SetPixel(xx,yy,c);
         } else if (!transparent) {
            SetPixel(xx,yy,bgc);
         }
         bit_mask = bit_mask >> 1;
      }

      work_char++;
   }
   return;
}

void Blit_Scaled_Character(int x, int y, unsigned char t, int c, int bgc,
                           bool transparent, float xscale, float yscale) {
   //Uses the ROM 8x8 character set to blit a character on the screen, scaled
   //to whatever size.

   unsigned char far *work_char;
   unsigned char far *start = ascii_text + (t << 3);
   unsigned char rs;
   unsigned char bit_mask = 0x80;
   float xx, yy, rh=CHAR_HEIGHT*yscale, rw=CHAR_WIDTH*xscale;

   //Compute starting offset into table.

   for (yy=y; yy < y + rh; yy++) {
      for (xx=x; xx < x + rw; xx++) {

         work_char = start + ((int)((yy-y)/rh*CHAR_HEIGHT));
         rs = (xx-x)/rw*CHAR_WIDTH;

         if (*work_char & bit_mask >> rs) {
            SetPixel(xx,yy,c);
         } else if (!transparent) {
            SetPixel(xx,yy,bgc);
         }
      }
   }
   return;
   
}

void Blit_String(char * string, int x, int y, int c, int bgc, bool trans) {

   int index=0;

   while (string[index] != '\0') {
      Blit_Character(x, y, string[index], c, bgc, trans);
      index++;
      x += CHAR_WIDTH;
   }
   return;
}

void Blit_Con_String(char * string, int &x, int y, int c, int bgc, bool trans) {

   int index=0;

   while (string[index] != '\0') {
      Blit_Character(x, y, string[index], c, bgc, trans);
      index++;
      x += CHAR_WIDTH;
   }
   return;
}

char * Input_String(int x, int y, int c, int bgc, int maxsize) {
   char * str = new char [maxsize+1];
   char a;
   int index=0;
   bool willtolive=true;

   while (willtolive) {
      while (!kbhit()) {}
      a=getch();
      switch (a) {
      case 0x0D:   //Return
         str[index] = '\0';
         willtolive=false;
         break;
      case 0x08:   //Backspace
         if (index > 0) {
            x-=CHAR_WIDTH;
            Blit_Character(x,y,str[index-1],bgc,bgc,true);
            index--;
         }
         break;
      default:    //Any other character
         str[index] = a;
         if (index < maxsize) {
            index++;
            Blit_Character(x, y, a, c, bgc, true);
            x+=CHAR_WIDTH;
         }
      }
   }
   return str;
}

char * Input_Con_String(int &x, int y, int c, int bgc, int maxsize,bool rot) {
   char * str = new char [maxsize+1];
   char a;
   int index=0;
   bool willtolive=true;
   RGB_color temp, temp2;

   while (willtolive) {
      if (!rot) {
         while (!kbhit()) {}
      } else {
         while(!kbhit()) {
            Rotate_Palette();
         }
      }
      a=getch();
      switch (a) {
      case 0x0D:   //Return
         str[index] = '\0';
         willtolive=false;
         break;
      case 0x08:   //Backspace
         if (index > 0) {
            x-=CHAR_WIDTH;
            Blit_Character(x,y,str[index-1],bgc,bgc,true);
            index--;
         }
         break;
      default:    //Any other character
         str[index] = a;
         if (index < maxsize) {
            index++;
            Blit_Character(x, y, a, c, bgc, true);
            x+=CHAR_WIDTH;
         }
      }
   }
   return str;
}

void CR(int &x, int &y) {
   x = 0;
   y += CHAR_HEIGHT;
   return;
}

void Blit_Scaled_String(char * string, int x, int y, int c, int bgc,
                        bool trans, float xscale, float yscale) {

   int index=0;
   float xx=x;

   while (string[index] != '\0') {
      Blit_Scaled_Character(xx, y, string[index], c, bgc, trans,
                            xscale, yscale);
      index++;
      xx += ((float)CHAR_WIDTH)*xscale;
   }
   return;
}

void Box(int minx, int miny, int maxx, int maxy, unsigned char color) {
   H_Line(minx, maxx, miny, color);
   H_Line(minx, maxx, maxy, color);
   V_Line(minx, miny, maxy, color);
   V_Line(maxx, miny, maxy, color);

   return;
}

void Filled_Box(int minx, int miny, int maxx, int maxy, unsigned char color) {
   int y;
   for (y=miny; y <= maxy; y++) {
      H_Line(minx, maxx, y, color);
   }
   return;
}
